@@ -4,14 +4,13 @@ mod error;
 mod global_state;
 mod module_state;
 
-pub use self::context::ModuleDependencyMap;
 use self::{
-    chain_map::ChainMap, context::Context, error::CompileError, global_state::GlobalState,
-    module_state::ModuleState,
+    chain_map::ChainMap, context::Context, global_state::GlobalState, module_state::ModuleState,
 };
+pub use self::{context::ModuleDependencyMap, error::CompileError};
 use crate::{
     ast,
-    ir::{Build, Configuration},
+    ir::{Build, Configuration, DynamicBuild, DynamicConfiguration, Rule},
 };
 use once_cell::sync::Lazy;
 use regex::{Captures, Regex};
@@ -21,6 +20,10 @@ use std::{
     sync::Arc,
 };
 
+const PHONY_RULE: &str = "phony";
+const BUILD_DIRECTORY_VARIABLE: &str = "builddir";
+const DYNAMIC_MODULE_VARIABLE: &str = "dyndep";
+
 static VARIABLE_PATTERN: Lazy<Regex> =
     Lazy::new(|| Regex::new(r"\$([[:alpha:]][[:alnum:]]*)").unwrap());
 
@@ -53,7 +56,14 @@ pub fn compile(
         global_state.default_outputs
     };
 
-    Ok(Configuration::new(global_state.outputs, default_outputs))
+    Ok(Configuration::new(
+        global_state.outputs,
+        default_outputs,
+        module_state
+            .variables
+            .get(BUILD_DIRECTORY_VARIABLE)
+            .cloned(),
+    ))
 }
 
 fn compile_module(
@@ -70,10 +80,6 @@ fn compile_module(
     for statement in module.statements() {
         match statement {
             ast::Statement::Build(build) => {
-                let rule = &module_state
-                    .rules
-                    .get(build.rule())
-                    .ok_or_else(|| CompileError::RuleNotFound(build.rule().into()))?;
                 let mut variables = module_state.variables.fork();
 
                 variables.extend(
@@ -89,15 +95,37 @@ fn compile_module(
 
                 let ir = Arc::new(Build::new(
                     context.generate_build_id(),
-                    interpolate_variables(rule.command(), &variables),
-                    interpolate_variables(rule.description(), &variables),
-                    build.inputs().to_vec(),
+                    build.outputs().to_vec(),
+                    build.implicit_outputs().to_vec(),
+                    if build.rule() == PHONY_RULE {
+                        None
+                    } else {
+                        let rule = &module_state
+                            .rules
+                            .get(build.rule())
+                            .ok_or_else(|| CompileError::RuleNotFound(build.rule().into()))?;
+
+                        Some(Rule::new(
+                            interpolate_variables(rule.command(), &variables),
+                            rule.description()
+                                .map(|description| interpolate_variables(description, &variables)),
+                        ))
+                    },
+                    build
+                        .inputs()
+                        .iter()
+                        .chain(build.implicit_inputs())
+                        .cloned()
+                        .collect(),
+                    build.order_only_inputs().to_vec(),
+                    variables.get(DYNAMIC_MODULE_VARIABLE).cloned(),
                 ));
 
                 global_state.outputs.extend(
                     build
                         .outputs()
                         .iter()
+                        .chain(build.implicit_outputs())
                         .map(|output| (output.clone(), ir.clone())),
                 );
             }
@@ -136,6 +164,21 @@ fn compile_module(
     Ok(())
 }
 
+pub fn compile_dynamic(module: &ast::DynamicModule) -> Result<DynamicConfiguration, CompileError> {
+    Ok(DynamicConfiguration::new(
+        module
+            .builds()
+            .iter()
+            .map(|build| {
+                (
+                    build.output().into(),
+                    DynamicBuild::new(build.implicit_inputs().to_vec()),
+                )
+            })
+            .collect(),
+    ))
+}
+
 fn resolve_dependency<'a>(
     context: &'a Context,
     module_path: &Path,
@@ -160,7 +203,7 @@ fn interpolate_variables(template: &str, variables: &ChainMap<String, String>) -
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::{ast, ir};
+    use crate::ast;
     use once_cell::sync::Lazy;
     use pretty_assertions::assert_eq;
 
@@ -171,6 +214,32 @@ mod tests {
             .collect()
     });
 
+    fn ast_explicit_build(
+        outputs: Vec<String>,
+        rule: impl Into<String>,
+        inputs: Vec<String>,
+        variable_definitions: Vec<ast::VariableDefinition>,
+    ) -> ast::Build {
+        ast::Build::new(
+            outputs,
+            vec![],
+            rule,
+            inputs,
+            vec![],
+            vec![],
+            variable_definitions,
+        )
+    }
+
+    fn ir_explicit_build(
+        id: impl Into<String>,
+        outputs: Vec<String>,
+        rule: Rule,
+        inputs: Vec<String>,
+    ) -> Build {
+        Build::new(id, outputs, vec![], rule.into(), inputs, vec![], None)
+    }
+
     #[test]
     fn compile_empty_module() {
         assert_eq!(
@@ -182,7 +251,7 @@ mod tests {
                 &ROOT_MODULE_PATH
             )
             .unwrap(),
-            ir::Configuration::new(Default::default(), Default::default())
+            Configuration::new(Default::default(), Default::default(), None)
         );
     }
 
@@ -194,8 +263,8 @@ mod tests {
                     ROOT_MODULE_PATH.clone(),
                     ast::Module::new(vec![
                         ast::VariableDefinition::new("x", "42").into(),
-                        ast::Rule::new("foo", "$x", "").into(),
-                        ast::Build::new(vec!["bar".into()], "foo", vec![], vec![]).into(),
+                        ast::Rule::new("foo", "$x", None).into(),
+                        ast_explicit_build(vec!["bar".into()], "foo", vec![], vec![]).into(),
                     ])
                 )]
                 .into_iter()
@@ -204,11 +273,16 @@ mod tests {
                 &ROOT_MODULE_PATH
             )
             .unwrap(),
-            ir::Configuration::new(
-                [("bar".into(), Build::new("0", "42", "", vec![]).into())]
-                    .into_iter()
-                    .collect(),
-                ["bar".into()].into_iter().collect()
+            Configuration::new(
+                [(
+                    "bar".into(),
+                    ir_explicit_build("0", vec!["bar".into()], Rule::new("42", None), vec![])
+                        .into()
+                )]
+                .into_iter()
+                .collect(),
+                ["bar".into()].into_iter().collect(),
+                None
             )
         );
     }
@@ -222,8 +296,8 @@ mod tests {
                     ast::Module::new(vec![
                         ast::VariableDefinition::new("x", "1").into(),
                         ast::VariableDefinition::new("y", "2").into(),
-                        ast::Rule::new("foo", "$x $y", "").into(),
-                        ast::Build::new(vec!["bar".into()], "foo", vec![], vec![]).into(),
+                        ast::Rule::new("foo", "$x $y", None).into(),
+                        ast_explicit_build(vec!["bar".into()], "foo", vec![], vec![]).into(),
                     ])
                 )]
                 .into_iter()
@@ -232,11 +306,16 @@ mod tests {
                 &ROOT_MODULE_PATH
             )
             .unwrap(),
-            ir::Configuration::new(
-                [("bar".into(), Build::new("0", "1 2", "", vec![]).into())]
-                    .into_iter()
-                    .collect(),
-                ["bar".into()].into_iter().collect()
+            Configuration::new(
+                [(
+                    "bar".into(),
+                    ir_explicit_build("0", vec!["bar".into()], Rule::new("1 2", None), vec![])
+                        .into()
+                )]
+                .into_iter()
+                .collect(),
+                ["bar".into()].into_iter().collect(),
+                None
             )
         );
     }
@@ -248,9 +327,8 @@ mod tests {
                 &[(
                     ROOT_MODULE_PATH.clone(),
                     ast::Module::new(vec![
-                        ast::VariableDefinition::new("x", "42").into(),
-                        ast::Rule::new("foo", "$$", "").into(),
-                        ast::Build::new(vec!["bar".into()], "foo", vec![], vec![]).into()
+                        ast::Rule::new("foo", "$$", None).into(),
+                        ast_explicit_build(vec!["bar".into()], "foo", vec![], vec![]).into()
                     ])
                 )]
                 .into_iter()
@@ -259,11 +337,15 @@ mod tests {
                 &ROOT_MODULE_PATH
             )
             .unwrap(),
-            ir::Configuration::new(
-                [("bar".into(), Build::new("0", "$", "", vec![]).into())]
-                    .into_iter()
-                    .collect(),
-                ["bar".into()].into_iter().collect()
+            Configuration::new(
+                [(
+                    "bar".into(),
+                    ir_explicit_build("0", vec!["bar".into()], Rule::new("$", None), vec![]).into()
+                )]
+                .into_iter()
+                .collect(),
+                ["bar".into()].into_iter().collect(),
+                None
             )
         );
     }
@@ -275,9 +357,8 @@ mod tests {
                 &[(
                     ROOT_MODULE_PATH.clone(),
                     ast::Module::new(vec![
-                        ast::VariableDefinition::new("x", "42").into(),
-                        ast::Rule::new("foo", "$in", "").into(),
-                        ast::Build::new(vec!["bar".into()], "foo", vec!["baz".into()], vec![])
+                        ast::Rule::new("foo", "$in", None).into(),
+                        ast_explicit_build(vec!["bar".into()], "foo", vec!["baz".into()], vec![])
                             .into(),
                     ])
                 )]
@@ -287,14 +368,66 @@ mod tests {
                 &ROOT_MODULE_PATH
             )
             .unwrap(),
-            ir::Configuration::new(
+            Configuration::new(
                 [(
                     "bar".into(),
-                    Build::new("0", "baz", "", vec!["baz".into()]).into()
+                    ir_explicit_build(
+                        "0",
+                        vec!["bar".into()],
+                        Rule::new("baz", None),
+                        vec!["baz".into()]
+                    )
+                    .into()
                 )]
                 .into_iter()
                 .collect(),
-                ["bar".into()].into_iter().collect()
+                ["bar".into()].into_iter().collect(),
+                None
+            )
+        );
+    }
+
+    #[test]
+    fn interpolate_in_variable_with_implicit_input() {
+        assert_eq!(
+            compile(
+                &[(
+                    ROOT_MODULE_PATH.clone(),
+                    ast::Module::new(vec![
+                        ast::Rule::new("foo", "$in", None).into(),
+                        ast::Build::new(
+                            vec!["bar".into()],
+                            vec![],
+                            "foo",
+                            vec!["baz".into()],
+                            vec!["blah".into()],
+                            vec![],
+                            vec![]
+                        )
+                        .into(),
+                    ])
+                )]
+                .into_iter()
+                .collect(),
+                &DEFAULT_DEPENDENCIES,
+                &ROOT_MODULE_PATH
+            )
+            .unwrap(),
+            Configuration::new(
+                [(
+                    "bar".into(),
+                    ir_explicit_build(
+                        "0",
+                        vec!["bar".into()],
+                        Rule::new("baz", None),
+                        vec!["baz".into(), "blah".into()]
+                    )
+                    .into()
+                )]
+                .into_iter()
+                .collect(),
+                ["bar".into()].into_iter().collect(),
+                None
             )
         );
     }
@@ -306,9 +439,8 @@ mod tests {
                 &[(
                     ROOT_MODULE_PATH.clone(),
                     ast::Module::new(vec![
-                        ast::VariableDefinition::new("x", "42").into(),
-                        ast::Rule::new("foo", "$out", "").into(),
-                        ast::Build::new(vec!["bar".into()], "foo", vec![], vec![]).into(),
+                        ast::Rule::new("foo", "$out", None).into(),
+                        ast_explicit_build(vec!["bar".into()], "foo", vec![], vec![]).into(),
                     ])
                 )]
                 .into_iter()
@@ -317,11 +449,110 @@ mod tests {
                 &ROOT_MODULE_PATH
             )
             .unwrap(),
-            ir::Configuration::new(
-                [("bar".into(), Build::new("0", "bar", "", vec![]).into())]
+            Configuration::new(
+                [(
+                    "bar".into(),
+                    ir_explicit_build("0", vec!["bar".into()], Rule::new("bar", None), vec![])
+                        .into()
+                )]
+                .into_iter()
+                .collect(),
+                ["bar".into()].into_iter().collect(),
+                None
+            )
+        );
+    }
+
+    #[test]
+    fn interpolate_out_variable_with_implicit_output() {
+        let build = Arc::new(Build::new(
+            "0",
+            vec!["bar".into()],
+            vec!["baz".into()],
+            Rule::new("bar", None).into(),
+            vec![],
+            vec![],
+            None,
+        ));
+
+        assert_eq!(
+            compile(
+                &[(
+                    ROOT_MODULE_PATH.clone(),
+                    ast::Module::new(vec![
+                        ast::Rule::new("foo", "$out", None).into(),
+                        ast::Build::new(
+                            vec!["bar".into()],
+                            vec!["baz".into()],
+                            "foo",
+                            vec![],
+                            vec![],
+                            vec![],
+                            vec![]
+                        )
+                        .into(),
+                    ])
+                )]
+                .into_iter()
+                .collect(),
+                &DEFAULT_DEPENDENCIES,
+                &ROOT_MODULE_PATH
+            )
+            .unwrap(),
+            Configuration::new(
+                [("baz".into(), build.clone()), ("bar".into(), build)]
                     .into_iter()
                     .collect(),
-                ["bar".into()].into_iter().collect()
+                ["baz".into(), "bar".into()].into_iter().collect(),
+                None
+            )
+        );
+    }
+
+    #[test]
+    fn compile_order_only_inputs() {
+        assert_eq!(
+            compile(
+                &[(
+                    ROOT_MODULE_PATH.clone(),
+                    ast::Module::new(vec![
+                        ast::Rule::new("foo", "$in", None).into(),
+                        ast::Build::new(
+                            vec!["bar".into()],
+                            vec![],
+                            "foo",
+                            vec![],
+                            vec![],
+                            vec!["baz".into()],
+                            vec![]
+                        )
+                        .into(),
+                    ])
+                )]
+                .into_iter()
+                .collect(),
+                &DEFAULT_DEPENDENCIES,
+                &ROOT_MODULE_PATH
+            )
+            .unwrap(),
+            Configuration::new(
+                [(
+                    "bar".into(),
+                    Build::new(
+                        "0",
+                        vec!["bar".into()],
+                        vec![],
+                        Some(Rule::new("", None)),
+                        vec![],
+                        vec!["baz".into()],
+                        None
+                    )
+                    .into()
+                )]
+                .into_iter()
+                .collect(),
+                ["bar".into()].into_iter().collect(),
+                None
             )
         );
     }
@@ -333,9 +564,9 @@ mod tests {
                 &[(
                     ROOT_MODULE_PATH.clone(),
                     ast::Module::new(vec![
-                        ast::Rule::new("foo", "", "").into(),
-                        ast::Build::new(vec!["bar".into()], "foo", vec![], vec![]).into(),
-                        ast::Build::new(vec!["baz".into()], "foo", vec![], vec![]).into()
+                        ast::Rule::new("foo", "", None).into(),
+                        ast_explicit_build(vec!["bar".into()], "foo", vec![], vec![]).into(),
+                        ast_explicit_build(vec!["baz".into()], "foo", vec![], vec![]).into()
                     ])
                 )]
                 .into_iter()
@@ -344,14 +575,23 @@ mod tests {
                 &ROOT_MODULE_PATH
             )
             .unwrap(),
-            ir::Configuration::new(
+            Configuration::new(
                 [
-                    ("bar".into(), Build::new("0", "", "", vec![]).into()),
-                    ("baz".into(), Build::new("1", "", "", vec![]).into())
+                    (
+                        "bar".into(),
+                        ir_explicit_build("0", vec!["bar".into()], Rule::new("", None), vec![])
+                            .into()
+                    ),
+                    (
+                        "baz".into(),
+                        ir_explicit_build("1", vec!["baz".into()], Rule::new("", None), vec![])
+                            .into()
+                    )
                 ]
                 .into_iter()
                 .collect(),
-                ["bar".into(), "baz".into()].into_iter().collect()
+                ["bar".into(), "baz".into()].into_iter().collect(),
+                None
             )
         );
     }
@@ -363,8 +603,8 @@ mod tests {
                 &[(
                     ROOT_MODULE_PATH.clone(),
                     ast::Module::new(vec![
-                        ast::Rule::new("foo", "$x", "").into(),
-                        ast::Build::new(
+                        ast::Rule::new("foo", "$x", None).into(),
+                        ast_explicit_build(
                             vec!["bar".into()],
                             "foo",
                             vec![],
@@ -379,11 +619,118 @@ mod tests {
                 &ROOT_MODULE_PATH
             )
             .unwrap(),
-            ir::Configuration::new(
-                [("bar".into(), Build::new("0", "42", "", vec![]).into())]
-                    .into_iter()
-                    .collect(),
-                ["bar".into()].into_iter().collect()
+            Configuration::new(
+                [(
+                    "bar".into(),
+                    ir_explicit_build("0", vec!["bar".into()], Rule::new("42", None), vec![])
+                        .into()
+                )]
+                .into_iter()
+                .collect(),
+                ["bar".into()].into_iter().collect(),
+                None
+            )
+        );
+    }
+
+    #[test]
+    fn compile_phony_rule() {
+        assert_eq!(
+            compile(
+                &[(
+                    ROOT_MODULE_PATH.clone(),
+                    ast::Module::new(vec![ast_explicit_build(
+                        vec!["foo".into()],
+                        "phony",
+                        vec!["bar".into()],
+                        vec![]
+                    )
+                    .into(),])
+                )]
+                .into_iter()
+                .collect(),
+                &DEFAULT_DEPENDENCIES,
+                &ROOT_MODULE_PATH
+            )
+            .unwrap(),
+            Configuration::new(
+                [(
+                    "foo".into(),
+                    Build::new(
+                        "0",
+                        vec!["foo".into()],
+                        vec![],
+                        None,
+                        vec!["bar".into()],
+                        vec![],
+                        None
+                    )
+                    .into()
+                )]
+                .into_iter()
+                .collect(),
+                ["foo".into()].into_iter().collect(),
+                None
+            )
+        );
+    }
+
+    #[test]
+    fn compile_build_directory() {
+        assert_eq!(
+            compile(
+                &[(
+                    ROOT_MODULE_PATH.clone(),
+                    ast::Module::new(vec![ast::VariableDefinition::new("builddir", "foo").into()])
+                )]
+                .into_iter()
+                .collect(),
+                &DEFAULT_DEPENDENCIES,
+                &ROOT_MODULE_PATH
+            )
+            .unwrap(),
+            Configuration::new(Default::default(), Default::default(), Some("foo".into()))
+        );
+    }
+
+    #[test]
+    fn compile_dynamic_module_variable() {
+        assert_eq!(
+            compile(
+                &[(
+                    ROOT_MODULE_PATH.clone(),
+                    ast::Module::new(vec![ast_explicit_build(
+                        vec!["foo".into()],
+                        "phony",
+                        vec![],
+                        vec![ast::VariableDefinition::new("dyndep", "bar")]
+                    )
+                    .into()])
+                )]
+                .into_iter()
+                .collect(),
+                &DEFAULT_DEPENDENCIES,
+                &ROOT_MODULE_PATH
+            )
+            .unwrap(),
+            Configuration::new(
+                [(
+                    "foo".into(),
+                    Build::new(
+                        "0",
+                        vec!["foo".into()],
+                        vec![],
+                        None,
+                        vec![],
+                        vec![],
+                        Some("bar".into())
+                    )
+                    .into()
+                )]
+                .into_iter()
+                .collect(),
+                ["foo".into()].into_iter().collect(),
+                None
             )
         );
     }
@@ -409,8 +756,9 @@ mod tests {
                         (
                             SUBMODULE_PATH.into(),
                             ast::Module::new(vec![
-                                ast::Rule::new("foo", "$x", "").into(),
-                                ast::Build::new(vec!["bar".into()], "foo", vec![], vec![]).into()
+                                ast::Rule::new("foo", "$x", None).into(),
+                                ast_explicit_build(vec!["bar".into()], "foo", vec![], vec![])
+                                    .into()
                             ])
                         )
                     ]
@@ -427,11 +775,16 @@ mod tests {
                     &ROOT_MODULE_PATH
                 )
                 .unwrap(),
-                ir::Configuration::new(
-                    [("bar".into(), Build::new("0", "42", "", vec![]).into())]
-                        .into_iter()
-                        .collect(),
-                    ["bar".into()].into_iter().collect()
+                Configuration::new(
+                    [(
+                        "bar".into(),
+                        ir_explicit_build("0", vec!["bar".into()], Rule::new("42", None), vec![])
+                            .into()
+                    )]
+                    .into_iter()
+                    .collect(),
+                    ["bar".into()].into_iter().collect(),
+                    None
                 )
             );
         }
@@ -447,13 +800,13 @@ mod tests {
                             ROOT_MODULE_PATH.clone(),
                             ast::Module::new(vec![
                                 ast::VariableDefinition::new("x", "42").into(),
-                                ast::Rule::new("foo", "$x", "").into(),
+                                ast::Rule::new("foo", "$x", None).into(),
                                 ast::Submodule::new(SUBMODULE_PATH).into(),
                             ])
                         ),
                         (
                             SUBMODULE_PATH.into(),
-                            ast::Module::new(vec![ast::Build::new(
+                            ast::Module::new(vec![ast_explicit_build(
                                 vec!["bar".into()],
                                 "foo",
                                 vec![],
@@ -475,11 +828,16 @@ mod tests {
                     &ROOT_MODULE_PATH
                 )
                 .unwrap(),
-                ir::Configuration::new(
-                    [("bar".into(), Build::new("0", "42", "", vec![]).into())]
-                        .into_iter()
-                        .collect(),
-                    ["bar".into()].into_iter().collect()
+                Configuration::new(
+                    [(
+                        "bar".into(),
+                        ir_explicit_build("0", vec!["bar".into()], Rule::new("42", None), vec![])
+                            .into()
+                    )]
+                    .into_iter()
+                    .collect(),
+                    ["bar".into()].into_iter().collect(),
+                    None
                 )
             );
         }
@@ -495,9 +853,10 @@ mod tests {
                             ROOT_MODULE_PATH.clone(),
                             ast::Module::new(vec![
                                 ast::VariableDefinition::new("x", "42").into(),
-                                ast::Rule::new("foo", "$x", "").into(),
+                                ast::Rule::new("foo", "$x", None).into(),
                                 ast::Submodule::new(SUBMODULE_PATH).into(),
-                                ast::Build::new(vec!["bar".into()], "foo", vec![], vec![]).into(),
+                                ast_explicit_build(vec!["bar".into()], "foo", vec![], vec![])
+                                    .into(),
                             ])
                         ),
                         (
@@ -518,11 +877,16 @@ mod tests {
                     &ROOT_MODULE_PATH
                 )
                 .unwrap(),
-                ir::Configuration::new(
-                    [("bar".into(), Build::new("0", "42", "", vec![]).into())]
-                        .into_iter()
-                        .collect(),
-                    ["bar".into()].into_iter().collect()
+                Configuration::new(
+                    [(
+                        "bar".into(),
+                        ir_explicit_build("0", vec!["bar".into()], Rule::new("42", None), vec![])
+                            .into()
+                    )]
+                    .into_iter()
+                    .collect(),
+                    ["bar".into()].into_iter().collect(),
+                    None,
                 )
             );
         }
