@@ -8,13 +8,19 @@ use std::{
 pub struct Configuration {
     outputs: HashMap<String, Arc<Build>>,
     default_outputs: HashSet<String>,
+    build_directory: Option<String>,
 }
 
 impl Configuration {
-    pub fn new(outputs: HashMap<String, Arc<Build>>, default_outputs: HashSet<String>) -> Self {
+    pub fn new(
+        outputs: HashMap<String, Arc<Build>>,
+        default_outputs: HashSet<String>,
+        build_directory: Option<String>,
+    ) -> Self {
         Self {
             outputs,
             default_outputs,
+            build_directory,
         }
     }
 
@@ -25,4 +31,8 @@ impl Configuration {
     pub fn default_outputs(&self) -> &HashSet<String> {
         &self.default_outputs
     }
+
+    pub fn build_directory(&self) -> Option<&str> {
+        self.build_directory.as_deref()
+    }
 }
