@@ -5,6 +5,8 @@ mod error;
 mod ir;
 mod parse;
 mod run;
+mod utilities;
+mod validation;
 
 use arguments::Arguments;
 use ast::{Module, Statement};
@@ -16,21 +18,22 @@ use parse::parse;
 use run::run;
 use std::{
     collections::HashMap,
+    env::set_current_dir,
     error::Error,
     path::{Path, PathBuf},
     process::exit,
 };
-use tokio::{
-    fs::{self, File},
-    io::AsyncReadExt,
-};
+
+use utilities::{canonicalize_path, read_file};
+use validation::{validate_configuration, validate_modules};
 
 const DEFAULT_BUILD_FILE: &str = "build.ninja";
+const DEFAULT_LOG_PREFIX: &str = "turtle: ";
 
 #[tokio::main]
 async fn main() {
     if let Err(error) = execute().await {
-        eprintln!("{}", error);
+        eprintln!("{}{}", DEFAULT_LOG_PREFIX, error);
 
         exit(1)
     }
@@ -39,28 +42,39 @@ async fn main() {
 async fn execute() -> Result<(), Box<dyn Error>> {
     let arguments = Arguments::parse();
 
+    if let Some(directory) = &arguments.directory {
+        set_current_dir(directory)?;
+    }
+
     let root_module_path =
         canonicalize_path(&arguments.file.as_deref().unwrap_or(DEFAULT_BUILD_FILE)).await?;
     let (modules, dependencies) = read_modules(&root_module_path).await?;
 
-    run(
-        &compile(&modules, &dependencies, &root_module_path)?,
-        root_module_path.parent().unwrap(),
-    )
-    .await?;
+    validate_modules(&dependencies)?;
+
+    let configuration = compile(&modules, &dependencies, &root_module_path)?;
+
+    validate_configuration(&configuration)?;
+
+    let build_directory = configuration
+        .build_directory()
+        .map(PathBuf::from)
+        .unwrap_or_else(|| root_module_path.parent().unwrap().into());
+
+    run(configuration, &build_directory, arguments.job_limit).await?;
 
     Ok(())
 }
 
 async fn read_modules(
     path: &Path,
-) -> Result<(HashMap<PathBuf, Module>, ModuleDependencyMap), Box<dyn Error>> {
+) -> Result<(HashMap<PathBuf, Module>, ModuleDependencyMap), InfrastructureError> {
     let mut paths = vec![canonicalize_path(path).await?];
     let mut modules = HashMap::new();
     let mut dependencies = HashMap::new();
 
     while let Some(path) = paths.pop() {
-        let module = read_module(&path).await?;
+        let module = parse(&read_file(&path).await?)?;
 
         let submodule_paths = try_join_all(
             module
@@ -96,22 +110,3 @@ async fn resolve_submodule_path(
         canonicalize_path(module_path.parent().unwrap().join(submodule_path)).await?,
     ))
 }
-
-async fn read_module(path: &Path) -> Result<Module, Box<dyn Error>> {
-    let mut source = "".into();
-
-    File::open(path)
-        .await
-        .map_err(|error| InfrastructureError::with_path(error, path))?
-        .read_to_string(&mut source)
-        .await
-        .map_err(|error| InfrastructureError::with_path(error, path))?;
-
-    Ok(parse(&source)?)
-}
-
-async fn canonicalize_path(path: impl AsRef<Path>) -> Result<PathBuf, InfrastructureError> {
-    fs::canonicalize(&path)
-        .await
-        .map_err(|error| InfrastructureError::with_path(error, path))
-}
